from dataclasses import dataclass
from ayradb.rest.http.response import Response
from ayradb.core.manage_record.unescape import *

FIELD_SEPARATOR = b'\x3b'
RECORD_SEPARATOR = b'\n'


class QueryError:
    TABLE_NOT_FOUND = 1
    FIELD_NOT_FOUND = 2
    GENERIC_ERROR = 100


ERROR_DICT = {
    'ERROR: TFA_MTKO_016: trans_h__manage_transaction_kick_off__sqlq:ERROR: [00400]:(00006): MTKOSQLQ_001.001: error in trans_h__manage_transaction_kick_off__sqlq__primary: ERROR: [00400]:(00006): MTKOSQLQPRI_013: could not decompose the sql command: ERROR: [00400]:(00006): DECOMPSQLCOM_060: error in sql_h__decompose_command__select: ERROR: [00400]:(00006): DECOMPSQLCOMSELECT_020.001.002: the table does not exist' : QueryError.TABLE_NOT_FOUND
}


@dataclass
class QueryResponse:

    success: bool
    content: list
    error_code: int
    _error_msg: str

    def __init__(self, success, error_code, content=None, _error_msg=""):
        self.success = success
        self.error_code = error_code
        self.content = content
        self._error_msg = _error_msg

    @staticmethod
    def from_http_response(res: Response):
        if res.status_code == 200:
            body: bytes = res.body
            records = body.split(RECORD_SEPARATOR)
            content = []
            if records[records.__len__()-1] == b'':
                # Remove empty b'' generated by the last record terminator
                records.pop(records.__len__()-1)
            for record in records:
                columns = record.split(FIELD_SEPARATOR)
                record = {}
                for cursor in range(0, columns.__len__(), 2):
                    # For every pair key-value in the record
                    # splitted body is organized as [key, value, key, value, ...]
                    field_key = columns[cursor]
                    field_value = unescape(columns[cursor + 1])
                    record[field_key.decode('utf-8')] = field_value
                content.append(record)
            return QueryResponse(True, 0, content=content)
        else:
            return QueryResponse(False, QueryError.GENERIC_ERROR, content=[])





from dataclasses import dataclass
from ayradb.rest.http.header import Header


@dataclass
class Response:
    status_code: int
    status_message: str
    headers: dict
    body: bytes
    chunked: bool
    last_chunk: bool

    # Position trackers
    header_buffer_seek_pos: int
    body_parsing_start_pos: int

    # Parsing flags
    flag_proto_ok: bool
    flag_status_code_ok: bool
    flag_status_msg_ok: bool
    flag_headers_ok: bool
    __flag_body_ok: bool

    _HTTP_PROTO_BYTES_SIZE = 8
    _HTTP_STATUS_CODE_FIRST_BYTE_IDX = 9
    _HTTP_STATUS_MSG_FIRST_BYTE_IDX = 13
    _HTTP_FIRST_LINE_SIZE = 14
    _HTTP_LINE_TERMINATOR = b'\x0d\x0a'

    _HTTP_STATUS_DICT = {
        # 1xx INFORMATIONAL
        b'\x31\x30\x30': 100,  # 100 - Continue
        b'\x31\x30\x31': 101,  # 101 - Switching protocols
        b'\x31\x30\x32': 102,  # 102 - Processing
        b'\x31\x30\x33': 103,  # 103 - Early hints
        b'\x31\x39\x38': 198,  # 198 - Cherrydata-provisional-update-servers-coord
        # 2xx SUCCESS
        b'\x32\x30\x30': 200,  # 200 - OK
        b'\x32\x30\x31': 201,  # 201 - Created
        b'\x32\x30\x32': 202,  # 202 - Accepted
        b'\x32\x30\x33': 203,  # 203 - Non-Authoritative Information
        b'\x32\x30\x34': 204,  # 204 - No Content
        b'\x32\x30\x35': 205,  # 205 - Reset Content
        b'\x32\x30\x36': 206,  # 206 - Partial Content
        b'\x32\x30\x37': 207,  # 207 - Multi status
        # 3xx REDIRECT
        # TODO: add missing
        b'\x33\x30\x30': 300,  # 300 - Multiple choices
        b'\x33\x30\x31': 301,  # 301 - Moved permanently
        # 4xx CLIENT ERRORs
        b'\x34\x30\x30': 400,  # 400 - Bad request
        b'\x34\x30\x31': 401,  # 401 - Unauthorized
        b'\x34\x30\x32': 402,  # 402 - Payment required
        b'\x34\x30\x33': 403,  # 403 - Forbidden
        b'\x34\x30\x34': 404,  # 404 - Not found
        b'\x34\x30\x35': 405,  # 405 - Method not allowed
        # 5xx CLIENT ERRORs
        b'\x35\x30\x30': 500,  # 500 - Internal Server Error
        b'\x35\x30\x31': 501,  # 501 - Not Implemented
        b'\x35\x30\x32': 502,  # 502 - Bad Gateway
        b'\x35\x30\x33': 503,  # 503 - Service Unavailable
        b'\x35\x30\x34': 504,  # 504 - Gateway Timeout
        b'\x35\x30\x35': 505,  # 505 - HTTP Version Not Supported
    }

    def __init__(self):
        self.flag_proto_ok = False
        self.flag_status_code_ok = False
        self.flag_status_msg_ok = False
        self.flag_headers_ok = False
        self.__flag_body_ok = False
        self.header_buffer_seek_pos = 0
        self.body_parsing_start_pos = 0
        self.headers = {}
        self.body = b''
        self.chunked = False
        self.last_chunk = False

    def from_byte_array(self, byte_array):
        if not self.flag_proto_ok:
            # HTTP proto still not parsed
            if byte_array.__len__() < Response._HTTP_PROTO_BYTES_SIZE:
                # Not enough bytes to check protocol
                return 0
            else:
                # Verify that in first 8 bytes there is HTTP/1.1
                self.flag_proto_ok = byte_array[
                                     0:Response._HTTP_PROTO_BYTES_SIZE] == b'\x48\x54\x54\x50\x2F\x31\x2e\x31'
                if not self.flag_proto_ok:
                    # Malformed message: HTTP/1.1 isn't in first 8 bytes
                    return -1

        if not self.flag_status_code_ok:
            # Status code not parsed

            # Status code still not parsed
            if byte_array.__len__() < Response._HTTP_FIRST_LINE_SIZE:
                # Not enough bytes to check status
                return 0
            # Check that a space is present after HTTP/1.1
            if byte_array[Response._HTTP_PROTO_BYTES_SIZE] != 0x20:
                # Case space not present (malformed message)
                return -1
            # Check http status code
            status_bytes = byte_array[
                           Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX:Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX + 3]
            self.status_code = Response._HTTP_STATUS_DICT.get(status_bytes)
            if self.status_code is None:
                # Bytes didn't correspond to a valid HTTP status
                return -1
            # Check that a space is present after the status
            if byte_array[Response._HTTP_STATUS_CODE_FIRST_BYTE_IDX + 3] != 0x20:
                # Space not present (malformed message)
                return -1
            # Status code parsed correctly
            self.flag_status_code_ok = True

        if not self.flag_status_msg_ok:
            # Status message not parsed

            # Start parsing from status message first byte position
            status_msg_pos = Response._HTTP_STATUS_MSG_FIRST_BYTE_IDX
            # Seek next \r\n
            next_rn_position = Response._seek_line_terminator(byte_array, start=status_msg_pos)
            if next_rn_position == -1:
                # \r\n not found (not enough bytes)
                return 0
            else:
                self.flag_status_msg_ok = True
                # Decode status msg
                self.status_message = byte_array[status_msg_pos:next_rn_position] \
                    .decode('utf8')
                # Start searching for the headers from the first byte
                # after the line terminator
                self.header_buffer_seek_pos = next_rn_position + Response._HTTP_LINE_TERMINATOR.__len__()

        while not self.flag_headers_ok:
            # Header still not parsed

            # Seek next \r\n starting from the first not parsed
            # byte of the headers
            next_rn_position = Response._seek_line_terminator(byte_array, start=self.header_buffer_seek_pos)
            if next_rn_position == -1:
                # \r\n not found (not enough bytes)
                return 0
            if next_rn_position == self.header_buffer_seek_pos:
                # This \r\n is preceded by another \r\n
                # Double line terminator means headers end
                # Since a HTTP message has at least 1 header (Content-Length)
                # we can skip the case with 0 headers
                self.flag_headers_ok = True
                self.body_parsing_start_pos = self.header_buffer_seek_pos + Response._HTTP_LINE_TERMINATOR.__len__()
                break
            # Decode header row
            header_row = byte_array[self.header_buffer_seek_pos:next_rn_position] \
                .decode('utf8')
            # Split header in key value
            splitted_header = header_row.split(':')
            if splitted_header[1][0] == ' ':
                # Remove whitespace before adding key value pair to headers
                self.headers[splitted_header[0]] = splitted_header[1][1:]
            else:
                # Add key value pair to headers
                self.headers[splitted_header[0]] = splitted_header[1]
            # Move header start position after \r\n
            self.header_buffer_seek_pos = next_rn_position + Response._HTTP_LINE_TERMINATOR.__len__()

        if self.headers.get(Header.TRANSFER_ENCODING) == 'chunked':
            self.chunked = True
            while not self.__flag_body_ok:
                next_rn_position = Response._seek_line_terminator(byte_array, start=self.body_parsing_start_pos)
                if next_rn_position == -1:
                    # \r\n not found (not enough bytes)
                    return 0
                chunk_start_pos = next_rn_position + Response._HTTP_LINE_TERMINATOR.__len__()
                if next_rn_position == self.body_parsing_start_pos:
                    # Case end of message (but not last chunk)
                    self.__flag_body_ok = True
                    self.body_parsing_start_pos += Response._HTTP_LINE_TERMINATOR.__len__()
                else:
                    # Retrieve chunk length and convert it to int
                    chunk_length = Response.__retrieve_chunk_length(byte_array, self.body_parsing_start_pos,
                                                                    next_rn_position)
                    if chunk_length == 0:
                        self.last_chunk = True
                        self.__flag_body_ok = True
                    else:
                        self.body += (byte_array[chunk_start_pos:chunk_start_pos + chunk_length])
                    # Update parsing start position for next iteration
                    self.body_parsing_start_pos = chunk_start_pos + chunk_length + Response._HTTP_LINE_TERMINATOR.__len__()
                    # Extract the record
            return self.body_parsing_start_pos

        else:
            # Get message length
            content_length = int(self.headers.get(Header.CONTENT_LENGTH))
            missing_bytes = content_length - self.body.__len__()
            parsable_bytes = 0
            if byte_array.__len__() < self.body_parsing_start_pos + missing_bytes:
                # Case not enough bytes to parse the whole body
                # Parse all the available bytes
                parsable_bytes = byte_array.__len__() - self.body_parsing_start_pos
            else:
                parsable_bytes = missing_bytes
            self.body += byte_array[self.body_parsing_start_pos:self.body_parsing_start_pos + parsable_bytes]
            self.body_parsing_start_pos = self.body_parsing_start_pos + parsable_bytes
            if self.body.__len__() != content_length:
                # Missing bytes from body
                return 0
            else:
                return self.body_parsing_start_pos

    @staticmethod
    def __retrieve_chunk_length(byte_array, start, end):
        # Retrieve chunk length and convert it to int
        chunk_length = byte_array[start:end].decode('ascii')
        return int(chunk_length, 16)  # The length is in HEX notation, that is why we insert 16.

    @staticmethod
    def _seek_line_terminator(byte_array, start=0):
        return Response._seek_sub_slice(byte_array, Response._HTTP_LINE_TERMINATOR, start=start)

    @staticmethod
    def _seek_sub_slice(byte_array, sub_slice, start=0):
        # Start searching from the start position
        idx = start
        sub_sl_len = sub_slice.__len__()
        # Search up for the first sub slice byte 
        # up to (byte_array.length - sub_slice.length) byte
        # After that we don't have enough byte to find the pattern
        while idx + sub_sl_len - 1 < byte_array.__len__():
            if byte_array[idx] == sub_slice[0]:
                # First byte of the pattern found
                idx += 1
                # Move to next and check remaining bytes
                if byte_array[idx:idx + sub_sl_len - 1] == sub_slice[1:]:
                    # Pattern found (Move back to original position)
                    return idx - 1
            else:
                idx += 1
        # Pattern not found
        return -1



__all__ = ['Promise']

import time
from dataclasses import dataclass
from threading import Lock

from ayradb.rest.http.request import Request
from ayradb.rest.http.response import Response
from ayradb.rest.stream import Stream

WAIT_SLEEP = 0.00005  # 50us


@dataclass
class Promise:
    __id_count = 0
    __lock = Lock()

    _req: Request
    informational_msgs: []

    def __init__(self, request, build: type(lambda x: Response)):
        self._req = request
        Promise.__lock.acquire()
        self.cID = Promise.__id_count
        Promise.__id_count += 1
        Promise.__lock.release()
        self.__stream = Stream(build)
        self.informational_msgs = []

    def get_request(self):
        return self._req

    def wait_response(self):
        while self.__stream.is_empty():
            time.sleep(WAIT_SLEEP)
        return self.__stream._get_first()

    def submit(self, response):
        self.__stream._submit(response)

    def _close(self):
        self.__stream._close()

    @property
    def stream(self):
        return self.__stream

    def submit_informational(self, response):
        self.informational_msgs.append(response)

    @staticmethod
    def reject(error_object):
        def error_function(error):
            return error
        promise = Promise(None, error_function)
        promise.__stream._submit(error_object)
        return promise



__all__ = ['Stream']

import time

from dataclasses import dataclass

from threading import Lock

from ayradb.rest.http.response import Response

WAIT_SLEEP = 0.00005  # 50us


class StreamClosedException(Exception):
    pass


@dataclass
class Stream:

    __res: list
    res_method: type(lambda x: Response)
    __lock: Lock
    __is_closed: bool

    def __init__(self, build: type(lambda x: Response)):
        self.__res = []
        self.res_method = build
        self.__lock = Lock()
        self.__is_closed = False

    def size(self):
        self.__lock.acquire()
        size = self.__res.__len__()
        self.__lock.release()
        return size

    def is_empty(self):
        return self.size() <= 0

    def take(self):
        self.__lock.acquire()
        elements = self.__res
        self.__res = []
        self.__lock.release()
        return elements

    def take_when_available(self, min_size=1):
        size = self.size()
        if self.is_closed() and size == 0:
            raise StreamClosedException
        while self.size() < min_size:
            time.sleep(WAIT_SLEEP)
        return self.take()

    def _get_first(self):
        self.__lock.acquire()
        first_el = self.__res[0]
        self.__lock.release()
        return first_el

    def _submit(self, response):
        self.__lock.acquire()
        self.__res.append(self.res_method(response))
        self.__lock.release()

    def _close(self):
        self.__lock.acquire()
        self.__is_closed = True
        self.__lock.release()

    def is_closed(self):
        self.__lock.acquire()
        is_closed = self.__is_closed
        self.__lock.release()
        return is_closed and self.is_empty()



from ayradb.rest.http.request import Request
from ayradb.rest.rest_manager import RestManager
from ayradb.rest.promise import Promise
from ayradb.core.endpoints.endpoints import SYSTEM_ENDPOINT
from ayradb.core.manage_record.query.response import QueryResponse

import json


def query_records(sql_query) -> Promise:
    # Create query
    query = {
        "action": "sqlq",
    }
    body = {
        "sql_query": sql_query,
    }
    query_request = Request(Request.METHOD_POST, path=SYSTEM_ENDPOINT, query=query,
                            body=json.dumps(body).encode('ascii'))

    return RestManager().submit_to_long_queue(query_request, QueryResponse.from_http_response)





from dataclasses import dataclass

from ayradb.rest.promise import Promise
from ayradb.core.manage_record.query import query_records


@dataclass
class Query:

    query_string: str

    def __init__(self, query_string):
        self.query_string = query_string

    def execute(self) -> Promise:
        return query_records(self.query_string)
